#!/usr/bin/env python

from neuron import h, gui
import numpy
import multiprocessing
import itertools as it
import run as rn
import time
import json
import os.path as osp


def attach_noise_sin_clamp(cell, delay, dur, offset, amp, freq, dt, tau, sigma, mu, loc, seed):
    """Attach a sinusoidal current Clamp to a cell.

    :param cell: Cell object to attach the current clamp.
    :param delay: Onset of the injected current.
    :param dur: Duration of the stimulus.
    :param offset: Offset of the sine.
    :param amp: The amplitude of the sine.
    :param freq: The frequency of the sine.
    :param dt: integration timestep.
    :param tau: Tau for the OU process
    :param sigma: The standard deviation of the normrand.
    :param mu: Mean of the normrand.
    :param loc: Location on the dendrite where the stimulus is placed.
    :param seed: Seed for the random number generation.
    """
    stim = h.IClampNoiseSin(cell.soma[0](loc))
    stim.delay = delay
    stim.dur = dur
    stim.std = sigma
    stim.offset = offset
    stim.amp = amp
    stim.freq = freq
    stim.dt = dt
    stim.tau = tau
    stim.mu = mu
    stim.new_seed = seed
    return stim


def simulate(simdur, dt, init, cel):
    """Start the simulation
    :param simdur: The simulation duration
    :param dt: The time step
    :param init: The initiation voltage
    :param cel: The temperature of the simulationn."""
    h.dt = dt
    h.tstop = simdur
    h.v_init = init
    h.celsius = cel
    h.run()


def writetofile(*arg):
    """Writing a variable number of variables to a txt file"""
    filename = arg[0]
    temparray = numpy.zeros((len(arg[1]), len(arg)-1))
    maxlength = len(arg)-1
    for i in range(maxlength):  # Loop over all the input arguments
        temp = numpy.array(arg[i + 1])
        temparray[:, i] = temp
    numpy.savetxt(filename, temparray)


def simulatepar(delay, dur, std, offset, amp, freq, dt, tau, mu, loc, simdur, init, cel, seed, loop, desMfr, synapse):
    """Simulating the model.
    
    :param delay: Time delay before the stimulation is applied (in ms).
    :param dur: The duration of the stimulation (in ms).
    :param std: The std of the input current.
    :param offset: The mean value of the input current.
    :param amp: The amplitude of the sine in the input current. (Not used in these simulations)
    :param freq: The frequency of the sine in the input current. (Not used in these simulations)
    :param dt: The integration timestep of the simulation.
    :param tau: Tau of the OU process.
    :param mu: Mean of the OU process.
    :param loc: The location where the current needs to be injected.
    :param simdur: The duration of the simulqtion (in ms).
    :param init: Starting membrane voltage of the simulation.
    :param cel: The temperature of the simulation.
    :param seed: The seed for the random number generator.
    :param loop: Unique identifier to denote which iteration of the simulation this is.
    :param desMfr: The required output firing rate of the model.
    :param synapse: Boolean variable to denote whether or not to include synapses."""

    h.cvode.cache_efficient(1)  # More efficient to set it 1

    # Create model and attach the current clamp
    cell = createModel(synapse)
    stim = attach_noise_sin_clamp(cell, delay, dur, offset, amp, freq, dt, tau, std, mu, loc, seed)

    # Setting up recording vectors
    somav = h.Vector()
    somav.record(cell.soma[0](0.5)._ref_v)
    iinj = h.Vector()
    iinj.record(stim._ref_i)

    # Setting up basic simulation parameters and running the simulation
    h.dt = dt
    h.tstop = simdur
    h.v_init = init
    h.celsius = cel
    h.run()

    # Saving the recorded data from the vectors
    filename = './OUTPUT/' + str(desMfr) + 'Hzmfr' + str(std) + 'std' + str(loop) + 'loop.txt'
    writetofile(filename, somav, iinj)

    # Making sure that the model is destroyed
    cell = None
    return 1


def simulatepar_one_arg(a):
    return simulatepar(*a)


def realmain(numprocs, initval, stdVal, desMfr, Erest, synapse):
    """Setting up the simulation to run in parallel.
    :param numprocs: The number of parallel processes to be used.
    :param initval: The mean of input current.
    :param stdVal: Std of the input current.
    :param desMfr: The desired output firing rate of the model.
    :param Erest: The resting membrane voltage.
    :param synapse: Boolean variable to denote whether or not to include synapses."""
    # Setting the number of processes to be used
    p = multiprocessing.Pool(processes=numprocs)

    # Initializing some basic variables
    dt = 0.005
    tau = 5
    mu = 0.
    loc = 0.5
    simdur = 60100.
    cel = 37
    init = Erest
    looprange = numpy.arange(1, numprocs+1, 1)
    amp = 0
    curr = initval
    delay = 100.
    dur = simdur
    seedrange = looprange
    freq = 0

    # Distributing the simulations to different processes and waiting until they are done
    q = p.map_async(simulatepar_one_arg, zip(it.repeat(delay), it.repeat(dur), it.repeat(stdVal), it.repeat(curr), it.repeat(amp), it.repeat(freq), it.repeat(dt), it.repeat(tau), it.repeat(mu), it.repeat(loc), it.repeat(simdur), it.repeat(init), it.repeat(cel), seedrange, looprange, it.repeat(desMfr), it.repeat(synapse)))
    while not q.ready():
        time.sleep(1)


def findStd(cell, desStd):
    """Using the bisection method to determine the correct value of std for the input current so that the std of
    the recorded membrane voltage reaches desStd.
    :param cell: The model that should be used.
    :param desStd: The required std of the membrane voltage."""
    # Setting some initial parameters for the current clamp
    delay = 0
    dur = 10000
    offset = 0
    amp = 0
    freq = 0
    dt = 0.005
    tau = 5
    sigma = 0
    mu = 0
    loc = 0.5
    seed = 1
    # Attaching the current clamp
    stim = attach_noise_sin_clamp(cell, delay, dur, offset, amp, freq, dt, tau, sigma, mu, loc, seed)

    # Setting up recording vectors
    somav = h.Vector()
    somav.record(cell.soma[0](0.5)._ref_v)
    timevec = h.Vector()
    timevec.record(h._ref_t)
    iinj = h.Vector()
    iinj.record(stim._ref_i)

    # Find input resistance and resting potential
    Erest, Rin = findInputRes(stim, somav, timevec, iinj)

    stim.offset = (-75-Erest)/Rin  # To have membrane potential at -75 mV

    # Find right std value of the noise with the bisection method
    h.v_init = Erest
    h.tstop = 2000
    stdVal = 0.001
    errorstd = 1  # To initialize the while loop
    stim.std = stdVal  # To initialize the while loop
    stdlow = 0
    stdhigh = 15/Rin
    somastdlow = 0
    somastdhigh = 30
    h.run()
    somaStd = somav.stdev(0, somav.size() - 1)
    somav.remove(0, somav.size() - 1)
    timevec.remove(0, timevec.size() - 1)
    iinj.remove(0, iinj.size() - 1)
    counter = 0
    while errorstd > 0.01:
        if desStd > somaStd:
            stdhigh = stdVal
            somastdhigh = somaStd
        else:
            stdlow = stdVal
            somastdlow = somaStd

        stdVal = (stdlow * (somastdhigh - desStd) - stdhigh * (somastdlow - desStd)) / (somastdhigh - somastdlow)
        stim.std = stdVal
        h.run()
        somaStd = somav.stdev(0, somav.size() - 1)
        errorstd = numpy.abs(desStd - somaStd)

        somav.remove(0, somav.size() - 1)
        timevec.remove(0, timevec.size() - 1)
        iinj.remove(0, iinj.size() - 1)
        counter += 1
        if counter > 500:  # Exit condition if it takes too many iterations to find a suitable solution
            exit("Couldn't find suitable std")

    strtoprint = "The input resistance is: " + str(Rin) + ", the resting potential is: " + str(Erest) + ", the standard deviation of the noise is set to " + str(stdVal) + "."
    print(strtoprint)
    return stdVal, Erest, Rin


def findInputRes(stim, somav, timevec, iinj):
    """Finding the input resistance of the model.
    :param stim: The current clamp.
    :param somav: A vector to hold soma membrane voltage recording.
    :param timevec: A vector to hold the time of the simulation.
    :param iinj: A vector to hold the injected current into the model"""
    # Performing a short simulation to make sure that the model is in the steady state
    tstop = 1000
    simulate(tstop, 0.005, -70, 37)
    Erest = somav.x[-2]
    h.tstop = 500
    Rin = numpy.zeros((11, 1))
    count = 0

    # Loop to go over hyperpolarizing step currents that are injected into the model and calculate the input resistance
    for i in numpy.arange(-0.02, -0.2, -0.18/10):
        stim.offset = i
        somav.remove(0, somav.size()-1)
        timevec.remove(0, timevec.size()-1)
        iinj.remove(0, iinj.size()-1)
        h.run()
        Rin[count] = (somav.x[-1]-Erest)/i
        count += 1
    somav.remove(0, somav.size() - 1)
    timevec.remove(0, timevec.size() - 1)
    iinj.remove(0, iinj.size() - 1)
    return Erest, numpy.mean(Rin)  # Returns the resting potential and input resistance


def findMfr(cell, stdVal, Erest, Rin, desMfr):
    """By using the bisection method this function will find the correct offset for the input current so that the
    output firing rate approximates desMfr.
    :param cell: The model.
    :param stdVal: The std of the input current.
    :param Erest: The resting membrane potential.
    :param Rin: The input resistance of the model.
    :param desMfr: The desired output firing rate of the model."""

    # Setting up an initial stimulator
    delay = 100
    dur = 1000000
    offset = 0.
    amp = 0
    freq = 0
    dt = 0.005
    tau = 5
    sigma = stdVal
    mu = 0
    loc = 0.5
    seed = 1
    stim = attach_noise_sin_clamp(cell, delay, dur, offset, amp, freq, dt, tau, sigma, mu, loc, seed)

    h.v_init = Erest
    h.tstop = 200000

    # Setting up the AP vector
    counts = h.Vector()
    apc = h.APCount(cell.soma[0](0.5))
    apc.thresh = 0
    apc.record(counts)

    # Loop parameters
    errormfr = 1
    offsetlow = -1
    offsethigh = (-Erest)/Rin
    mfrlow = 0
    mfrhigh = desMfr + 10
    h.run()
    if len(counts) > 2:
        currmfr = 1000 * len(counts) / (counts[-1] - counts[0])
        counts.remove(0, counts.size() - 1)
    else:
        currmfr = 0
    counter = 0
    # Using the bisection method to find the correct mean value of the input current to achieve a firing rate of desMfr
    while errormfr > 0.2:
        if currmfr > desMfr:
            offsethigh = offset
            mfrhigh = currmfr
        else:
            offsetlow = offset
            mfrlow = currmfr

        offset = (offsetlow * (mfrhigh - desMfr) - offsethigh * (mfrlow - desMfr)) / (mfrhigh - mfrlow)
        stim.offset = offset
        h.run()
        if len(counts) > 2:
            currmfr = 1000*len(counts)/(counts[-1]-counts[0])
            counts.remove(0, counts.size() - 1)
        else:
            currmfr = 0

        errormfr = abs(desMfr - currmfr)
        counter += 1
        if counter > 50:  # Exit condition when it takes too many iteration to find a correct value
            exit("Couldn't find a suitable offset for the desired mfr.")

    strtoprint = "The mean of the input is set to: " + str(offset) + "."
    print(strtoprint)
    return offset


def jsondump(Erest, Rin, stdVal, somastd, offset, currmfr, filenam):
    """Putting the respective calculated variables in a json file and saving it.
    :param Erest: Resting potential.
    :param Rin: Input resistance.
    :param stdVal: Std of the input current required to achieve a std value of somaStd in the output membrane voltage.
    :param somastd: The required std of the output membrane voltage.
    :param offset: The mean of the input current to achieve a firing rate of currmfr.
    :param currmfr: The required firing rate of the model.
    :param filenam: The filename in which all the data should be saved."""
    parameters = {"Erest": Erest,
                  "Rin": Rin,
                  "stdVal": stdVal,
                  "somaStd": somastd,
                  "offset": offset,
                  "mfr": currmfr,
                  }
    fid = open(filenam, "w")
    json.dump(parameters, fid)
    fid.close()


def runStdMfrProtocol(cell, desStd, desMfr):
    """Function that will find the desired std and mean for the input current given a required std of the membrane
    voltage and firing rate of the model. Saving the found variables in a json file.
    :param cell: The model that needs to be used.
    :param desStd: The desired std of the output membranve voltage.
    :param desMfr: The desired firing rate of the model."""

    filenam = "./Parameters/" + str(desStd) + "SomaStd" + str(desMfr) + "Hz.json"

    if osp.isfile(filenam):
        fid = open(filenam, "r")
        tempJsonVar = json.load(fid)
        fid.close()

        Erest = tempJsonVar["Erest"]
        Rin = tempJsonVar["Rin"]
        stdVal = tempJsonVar["stdVal"]
        somaStd = tempJsonVar["somaStd"]
        offset = tempJsonVar["offset"]
        mfr = tempJsonVar["mfr"]
        print("Found existing JSON file for the requested paramters. Opening existing JSON file and using these input "
              "parameters: \n - Erest: " + str(Erest) + "\n - Rin:" + str(Rin) + "\n - stdVal: " + str(stdVal) + "\n "
               "- somaStd: " + str(somaStd) + "\n - offset: " + str(offset) + "\n - mfr: " + str(mfr))
    else:
        print("No JSON file detected for required parameters, running the following protocol: \n - Mcm.findInputRes "
              "\n - Mcm.findStd \n - Mcm.findMfr \n - Mcm.jsondump")

        # Finding the right std for the noise
        stdVal, Erest, Rin = findStd(cell, desStd)

        # Finding the correct offset for the desired mfr
        offset = findMfr(cell, stdVal, Erest, Rin, desMfr)

        # Saving the found parameters in a json file
        jsondump(Erest, Rin, stdVal, desStd, offset, desMfr, filenam)

    return Erest, Rin, stdVal, offset


def createModel(synapse):
    """Creates the model with the basic implemented BBP functions.
    :param synapse: Boolean variable to denote whether or not to include synapses."""
    rn.init_simulation()
    cell = rn.create_cell(synapse)

    return cell
